#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: validate

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    7/10/18
"""
import daiquiri

from metapype.eml import rule
from metapype.eml.exceptions import MetapypeRuleError, UnknownNodeError
from metapype.eml.validation_errors import ValidationError
from metapype.model.node import Node


logger = daiquiri.getLogger("validate: " + __name__)


def node(n: Node, errs: list = None) -> None:
    """
    Validates a given node for rule compliance.

    Args:
        n: Node instance to be validated
        errs: List container for validation errors (fail fast if None)

    Returns:
        None

    Raises:
        MetapypeRuleError: An unknown type of node for EML
    """
    if n.name not in rule.node_mappings:
        msg = f"Unknown node rule type: {n.name}"
        if errs is None:
            raise UnknownNodeError(msg)
        else:
            errs.append((ValidationError.UNKNOWN_NODE, msg, n))
    else:
        node_rule = rule.get_rule(n.name)
        node_rule.validate_rule(n, errs)


def prune(n: Node) -> list:
    """
    Prune in place all non-valid nodes from the tree

    Args:
        n: Node

    Returns: List of pruned nodes

    Side-effects: Non-valid nodes are pruned from the tree

    """
    pruned = list()
    if n.name != "metadata":
        try:
            node(n)
        except MetapypeRuleError:
            for child in n.children:
                try:
                    node(child)
                except UnknownNodeError:
                    pruned.append(child)
            for child in pruned:
                n.remove_child(child)
                Node.delete_node_instance(child.id)
        for child in n.children:
            pruned += prune(child)
    return pruned


def tree(n: Node, errs: list = None) -> None:
    """
    Recursively walks from the root node and validates
    each child node for rule compliance.

    Args:
        n: Node instance of root for validates
        errs: List container for validation errors (fail fast if None)

    Returns:
        None
    """
    node(n, errs)
    if n.name != "metadata":
        for child in n.children:
            tree(child, errs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: evaluate.py

:Synopsis:

:Author:
    ide

:Created:
    6/24/20
"""
from enum import Enum, auto


class EvaluationWarning(Enum):
    CONNECTION_DEFINITION_DESCRIPTION_MISSING = auto()
    DATASET_ABSTRACT_MISSING = auto()
    DATASET_ABSTRACT_TOO_SHORT = auto()
    DATASET_COVERAGE_MISSING = auto()
    DATASET_METHOD_STEPS_MISSING = auto()
    DATASET_PROJECT_MISSING = auto()
    DATATABLE_DESCRIPTION_MISSING = auto()
    DATATABLE_MISSING = auto()
    DESIGN_DESCRIPTION_DESCRIPTION_MISSING = auto()
    EMAIL_MISSING = auto()
    INDIVIDUAL_NAME_INCOMPLETE = auto()
    INTELLECTUAL_RIGHTS_MISSING = auto()
    KEYWORDS_MISSING = auto()
    KEYWORDS_INSUFFICIENT = auto()
    MAINTENANCE_DESCRIPTION_MISSING = auto()
    METHOD_STEP_DESCRIPTION_MISSING = auto()
    ORCID_ID_MISSING = auto()
    OTHER_ENTITY_DESCRIPTION_MISSING = auto()
    PROCEDURE_STEP_DESCRIPTION_MISSING = auto()
    QUALITY_CONTROL_DESCRIPTION_MISSING = auto()
    SAMPLING_DESCRIPTION_DESCRIPTION_MISSING = auto()
    STUDY_EXTENT_DESCRIPTION_MISSING = auto()
    TITLE_TOO_SHORT = auto()
    USER_ID_MISSING = auto()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: harness

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    6/5/18
"""
import json
import logging
import os

import daiquiri

from metapype.eml.exceptions import MetapypeRuleError
from metapype.eml import export
from metapype.eml import evaluate
from metapype.eml import names
from metapype.eml import rule
from metapype.eml import validate
from metapype.model import mp_io
from metapype.model import metapype_io
from metapype.model.node import Node

cwd = os.path.dirname(os.path.realpath(__file__))
logfile = cwd + "/harness.log"
daiquiri.setup(
    level=logging.INFO, outputs=(daiquiri.output.File(logfile), "stdout",)
)
logger = daiquiri.getLogger(__name__)


def main():

    eml = Node(names.EML)
    eml.add_attribute("packageId", "edi.23.1")
    eml.add_attribute("system", "metapype")

    access = Node(names.ACCESS, parent=eml)
    access.add_attribute("authSystem", "pasta")
    access.add_attribute("order", "allowFirst")
    eml.add_child(access)

    allow = Node(names.ALLOW, parent=access)
    access.add_child(allow)

    principal = Node(names.PRINCIPAL, parent=allow)
    principal.content = "uid=gaucho,o=EDI,dc=edirepository,dc=org"
    allow.add_child(principal)

    permission = Node(names.PERMISSION, parent=allow)
    permission.content = "all"
    allow.add_child(permission)

    dataset = Node(names.DATASET, parent=eml)
    eml.add_child(dataset)

    title = Node(names.TITLE, parent=dataset)
    title.content = "Green sea turtle counts: Tortuga Island 20017"
    dataset.add_child(title)

    creator = Node(names.CREATOR, parent=dataset)
    dataset.add_child(creator)

    individualName_creator = Node(names.INDIVIDUALNAME, parent=creator)
    creator.add_child(individualName_creator)

    surName_creator = Node(names.SURNAME, parent=individualName_creator)
    surName_creator.content = "Gaucho"
    individualName_creator.add_child(surName_creator)

    pubdate = Node(names.PUBDATE, parent=dataset)
    pubdate.content = "2018"
    dataset.add_child(pubdate)

    abstract = Node(names.ABSTRACT, parent=dataset)
    dataset.add_child(abstract)

    section_1 = Node(names.SECTION, parent=abstract)
    abstract.add_child(section_1)

    section_2 = Node(names.SECTION, parent=abstract)
    abstract.add_child(section_2)

    para_1 = Node(names.PARA, parent=section_1)
    section_1.add_child(para_1)
    para_1.content = "This is abstract Para 1"

    para_2 = Node(names.PARA, parent=section_2)
    section_2.add_child(para_2)
    para_2.content = "This is abstract Para 2"

    intellectualRights = Node(names.INTELLECTUALRIGHTS, parent=dataset)
    dataset.add_child(intellectualRights)

    section_3 = Node(names.SECTION, parent=intellectualRights)
    intellectualRights.add_child(section_3)

    section_4 = Node(names.SECTION, parent=intellectualRights)
    intellectualRights.add_child(section_4)

    para_3 = Node(names.PARA, parent=section_3)
    section_3.add_child(para_3)
    para_3.content = "This is intellectualRights Para 3"

    para_4 = Node(names.PARA, parent=section_4)
    section_4.add_child(para_4)
    para_4.content = "This is intellectualRights Para 4"

    coverage = Node(names.COVERAGE, parent=dataset)
    dataset.add_child(coverage)
    geographic_coverage = Node(names.GEOGRAPHICCOVERAGE, parent=coverage)
    coverage.add_child(geographic_coverage)
    geographic_description = Node(
        names.GEOGRAPHICDESCRIPTION, parent=geographic_coverage
    )
    geographic_description.content = "Somewhere in the Rocky Mountains"
    geographic_coverage.add_child(geographic_description)
    bounding_coordinates = Node(
        names.BOUNDINGCOORDINATES, parent=geographic_coverage
    )
    geographic_coverage.add_child(bounding_coordinates)
    wbc = Node(names.WESTBOUNDINGCOORDINATE, parent=bounding_coordinates)
    wbc.content = -107.55538624999997
    bounding_coordinates.add_child(wbc)
    ebc = Node(names.EASTBOUNDINGCOORDINATE, parent=bounding_coordinates)
    ebc.content = -106.45675343749997
    bounding_coordinates.add_child(ebc)
    nbc = Node(names.NORTHBOUNDINGCOORDINATE, parent=bounding_coordinates)
    nbc.content = 39.48496522541802
    bounding_coordinates.add_child(nbc)
    sbc = Node(names.SOUTHBOUNDINGCOORDINATE, parent=bounding_coordinates)
    sbc.content = 38.9530013453599
    bounding_coordinates.add_child(sbc)

    temporal_coverage = Node(names.TEMPORALCOVERAGE, parent=coverage)
    coverage.add_child(temporal_coverage)

    # single_date_time = Node(names.SINGLEDATETIME, parent=temporal_coverage)
    # temporal_coverage.add_child(single_date_time)
    # single_date_time.content = "2018"

    range_of_dates = Node(names.RANGEOFDATES, parent=temporal_coverage)
    temporal_coverage.add_child(range_of_dates)

    begin_date = Node(names.BEGINDATE, parent=range_of_dates)
    range_of_dates.add_child(begin_date)

    calendar_date_begin = Node(names.CALENDARDATE, parent=begin_date)
    begin_date.add_child(calendar_date_begin)
    calendar_date_begin.content = "2000"

    end_date = Node(names.ENDDATE, parent=range_of_dates)
    range_of_dates.add_child(end_date)

    calendar_date_end = Node(names.CALENDARDATE, parent=end_date)
    end_date.add_child(calendar_date_end)
    calendar_date_end.content = "2018"

    temporal_coverage_rule = rule.get_rule(names.TEMPORALCOVERAGE)
    try:
        temporal_coverage_rule.validate_rule(temporal_coverage)
    except MetapypeRuleError as e:
        # We should see an error like:
        #   Maximum occurrence of "['singleDateTime', 'rangeOfDates']"
        #   exceeded for "temporalCoverage"
        logger.error(e)

    taxonomic_coverage = Node(names.TAXONOMICCOVERAGE, parent=coverage)
    coverage.add_child(taxonomic_coverage)
    general_taxonomic_coverage = Node(
        names.GENERALTAXONOMICCOVERAGE, parent=taxonomic_coverage
    )
    taxonomic_coverage.add_child(general_taxonomic_coverage)
    general_taxonomic_coverage.content = "All vascular plants were \
identified to family or species, mosses and lichens were \
identified as moss or lichen."

    taxonomic_classification_genus = Node(
        names.TAXONOMICCLASSIFICATION, parent=taxonomic_coverage
    )
    taxonomic_coverage.add_child(taxonomic_classification_genus)

    taxon_rank_name_genus = Node(
        names.TAXONRANKNAME, parent=taxonomic_classification_genus
    )
    taxonomic_classification_genus.add_child(taxon_rank_name_genus)
    taxon_rank_name_genus.content = "Genus"

    taxon_rank_value_genus = Node(
        names.TAXONRANKVALUE, parent=taxonomic_classification_genus
    )
    taxonomic_classification_genus.add_child(taxon_rank_value_genus)
    taxon_rank_value_genus.content = "Escherichia"

    taxonomic_classification_species = Node(
        names.TAXONOMICCLASSIFICATION, parent=taxonomic_classification_genus
    )
    taxonomic_classification_genus.add_child(taxonomic_classification_species)

    taxon_rank_name_species = Node(
        names.TAXONRANKNAME, parent=taxonomic_classification_species
    )
    taxonomic_classification_species.add_child(taxon_rank_name_species)
    taxon_rank_name_species.content = "Species"

    taxon_rank_value_species = Node(
        names.TAXONRANKVALUE, parent=taxonomic_classification_species
    )
    taxonomic_classification_species.add_child(taxon_rank_value_species)
    taxon_rank_value_species.content = "coli"

    contact = Node(names.CONTACT, parent=dataset)
    dataset.add_child(contact)

    individualName_contact = Node(names.INDIVIDUALNAME, parent=contact)
    contact.add_child(individualName_contact)

    givenName_contact = Node(names.GIVENNAME, parent=individualName_contact)
    givenName_contact.content = "Chase"
    individualName_contact.add_child(givenName_contact)

    surName_contact = Node(names.SURNAME, parent=individualName_contact)
    surName_contact.content = "Gaucho"
    individualName_contact.add_child(surName_contact)

    publisher = Node(names.PUBLISHER, parent=dataset)
    dataset.add_child(publisher)

    individualName_publisher = Node(names.INDIVIDUALNAME, parent=publisher)
    publisher.add_child(individualName_publisher)

    givenName_publisher = Node(
        names.GIVENNAME, parent=individualName_publisher
    )
    givenName_publisher.content = "Utah"
    individualName_publisher.add_child(givenName_publisher)

    surName_publisher = Node(names.SURNAME, parent=individualName_publisher)
    surName_publisher.content = "Carroll"
    individualName_publisher.add_child(surName_publisher)

    pubplace = Node(names.PUBPLACE, parent=dataset)
    dataset.add_child(pubplace)
    pubplace.content = "University of New Mexico Dept. of Biology"

    methods = Node(names.METHODS, parent=dataset)
    dataset.add_child(methods)

    method_step = Node(names.METHODSTEP, parent=methods)
    methods.add_child(method_step)

    description = Node(names.DESCRIPTION, parent=method_step)
    method_step.add_child(description)

    section_1 = Node(names.SECTION, parent=description)
    description.add_child(section_1)

    section_2 = Node(names.SECTION, parent=description)
    description.add_child(section_2)

    para_1 = Node(names.PARA, parent=section_1)
    section_1.add_child(para_1)
    para_1.content = "This is description Para 1"

    para_2 = Node(names.PARA, parent=section_2)
    section_2.add_child(para_2)
    para_2.content = "This is description Para 2"

    instrumentation_1 = Node(names.INSTRUMENTATION, parent=method_step)
    method_step.add_child(instrumentation_1)
    instrumentation_1.content = "Instrument #1"

    instrumentation_2 = Node(names.INSTRUMENTATION, parent=method_step)
    method_step.add_child(instrumentation_2)
    instrumentation_2.content = "Instrument #2"

    sampling = Node(names.SAMPLING, parent=methods)
    methods.add_child(sampling)

    study_extent = Node(names.STUDYEXTENT, parent=sampling)
    sampling.add_child(study_extent)

    study_extent_coverage = Node(names.COVERAGE, parent=study_extent)
    study_extent.add_child(study_extent_coverage)

    study_extent_geographic_coverage = Node(
        names.GEOGRAPHICCOVERAGE, parent=study_extent_coverage
    )
    study_extent_coverage.add_child(study_extent_geographic_coverage)

    study_extent_geographic_description = Node(
        names.GEOGRAPHICDESCRIPTION, parent=study_extent_geographic_coverage
    )
    study_extent_geographic_description.content = (
        "Somewhere in the Rocky Mountains"
    )
    study_extent_geographic_coverage.add_child(
        study_extent_geographic_description
    )

    study_extent_bounding_coordinates = Node(
        names.BOUNDINGCOORDINATES, parent=study_extent_geographic_coverage
    )
    study_extent_geographic_coverage.add_child(
        study_extent_bounding_coordinates
    )

    study_extent_wbc = Node(
        names.WESTBOUNDINGCOORDINATE, parent=study_extent_bounding_coordinates
    )
    study_extent_wbc.content = -107.55538624999997
    study_extent_bounding_coordinates.add_child(study_extent_wbc)
    study_extent_ebc = Node(
        names.EASTBOUNDINGCOORDINATE, parent=study_extent_bounding_coordinates
    )
    study_extent_ebc.content = -106.45675343749997
    study_extent_bounding_coordinates.add_child(study_extent_ebc)
    study_extent_nbc = Node(
        names.NORTHBOUNDINGCOORDINATE, parent=study_extent_bounding_coordinates
    )
    study_extent_nbc.content = 39.48496522541802
    study_extent_bounding_coordinates.add_child(study_extent_nbc)
    study_extent_sbc = Node(
        names.SOUTHBOUNDINGCOORDINATE, parent=study_extent_bounding_coordinates
    )
    study_extent_sbc.content = 38.9530013453599
    study_extent_bounding_coordinates.add_child(study_extent_sbc)

    study_extent_temporal_coverage = Node(
        names.TEMPORALCOVERAGE, parent=study_extent_coverage
    )
    study_extent_coverage.add_child(study_extent_temporal_coverage)

    study_extent_single_date_time = Node(
        names.SINGLEDATETIME, parent=study_extent_temporal_coverage
    )
    study_extent_temporal_coverage.add_child(study_extent_single_date_time)

    study_extent_calendar_date = Node(names.CALENDARDATE)
    study_extent_calendar_date.content = "2018"
    study_extent_single_date_time.add_child(study_extent_calendar_date)
    study_extent_time = Node(names.TIME)
    study_extent_time.content = "14:34:32.001-06:00"
    study_extent_single_date_time.add_child(study_extent_time)

    study_extent_range_of_dates = Node(
        names.RANGEOFDATES, parent=study_extent_temporal_coverage
    )
    # study_extent_temporal_coverage.add_child(study_extent_range_of_dates)

    study_extent_begin_date = Node(
        names.BEGINDATE, parent=study_extent_range_of_dates
    )
    study_extent_range_of_dates.add_child(study_extent_begin_date)

    study_extent_calendar_date_begin = Node(
        names.CALENDARDATE, parent=study_extent_begin_date
    )
    study_extent_begin_date.add_child(study_extent_calendar_date_begin)
    study_extent_calendar_date_begin.content = "2000"

    study_extent_end_date = Node(
        names.ENDDATE, parent=study_extent_range_of_dates
    )
    study_extent_range_of_dates.add_child(study_extent_end_date)

    study_extent_calendar_date_end = Node(
        names.CALENDARDATE, parent=study_extent_end_date
    )
    study_extent_end_date.add_child(study_extent_calendar_date_end)
    study_extent_calendar_date_end.content = "2018"

    study_extent_temporal_coverage_rule = rule.get_rule(names.TEMPORALCOVERAGE)
    try:
        study_extent_temporal_coverage_rule.validate_rule(temporal_coverage)
    except MetapypeRuleError as e:
        # We should see an error like:
        #   Maximum occurrence of "['singleDateTime', 'rangeOfDates']"
        #   exceeded for "temporalCoverage"
        logger.error(e)

    study_extent_taxonomic_coverage = Node(
        names.TAXONOMICCOVERAGE, parent=study_extent_coverage
    )
    study_extent_coverage.add_child(study_extent_taxonomic_coverage)
    study_extent_general_taxonomic_coverage = Node(
        names.GENERALTAXONOMICCOVERAGE, parent=study_extent_taxonomic_coverage
    )
    study_extent_taxonomic_coverage.add_child(
        study_extent_general_taxonomic_coverage
    )
    study_extent_general_taxonomic_coverage.content = "All vascular plants were \
identified to family or species, mosses and lichens were \
identified as moss or lichen."

    study_extent_taxonomic_classification_genus = Node(
        names.TAXONOMICCLASSIFICATION, parent=study_extent_taxonomic_coverage
    )
    study_extent_taxonomic_coverage.add_child(
        study_extent_taxonomic_classification_genus
    )

    study_extent_taxon_rank_name_genus = Node(
        names.TAXONRANKNAME, parent=study_extent_taxonomic_classification_genus
    )
    study_extent_taxonomic_classification_genus.add_child(
        study_extent_taxon_rank_name_genus
    )
    study_extent_taxon_rank_name_genus.content = "Genus"

    study_extent_taxon_rank_value_genus = Node(
        names.TAXONRANKVALUE,
        parent=study_extent_taxonomic_classification_genus,
    )
    study_extent_taxonomic_classification_genus.add_child(
        study_extent_taxon_rank_value_genus
    )
    study_extent_taxon_rank_value_genus.content = "Escherichia"

    study_extent_taxonomic_classification_species = Node(
        names.TAXONOMICCLASSIFICATION,
        parent=study_extent_taxonomic_classification_genus,
    )
    study_extent_taxonomic_classification_genus.add_child(
        study_extent_taxonomic_classification_species
    )

    study_extent_taxon_rank_name_species = Node(
        names.TAXONRANKNAME,
        parent=study_extent_taxonomic_classification_species,
    )
    study_extent_taxonomic_classification_species.add_child(
        study_extent_taxon_rank_name_species
    )
    study_extent_taxon_rank_name_species.content = "Species"

    study_extent_taxon_rank_value_species = Node(
        names.TAXONRANKVALUE,
        parent=study_extent_taxonomic_classification_species,
    )
    study_extent_taxonomic_classification_species.add_child(
        study_extent_taxon_rank_value_species
    )
    study_extent_taxon_rank_value_species.content = "coli"

    study_extent_description = Node(names.DESCRIPTION, parent=study_extent)
    study_extent.add_child(study_extent_description)

    study_extent_section_1 = Node(
        names.SECTION, parent=study_extent_description
    )
    study_extent_description.add_child(study_extent_section_1)

    study_extent_section_2 = Node(
        names.SECTION, parent=study_extent_description
    )
    study_extent_description.add_child(study_extent_section_2)

    study_extent_para_1 = Node(names.PARA, parent=study_extent_section_1)
    study_extent_section_1.add_child(study_extent_para_1)
    study_extent_para_1.content = "This is study_extent_description Para 1"

    study_extent_para_2 = Node(names.PARA, parent=study_extent_section_2)
    study_extent_section_2.add_child(study_extent_para_2)
    study_extent_para_2.content = "This is study_extent_description Para 2"

    sampling_description = Node(names.SAMPLINGDESCRIPTION, parent=sampling)
    sampling.add_child(sampling_description)

    sampling_description_section_1 = Node(
        names.SECTION, parent=sampling_description
    )
    sampling_description.add_child(sampling_description_section_1)

    sampling_description_section_2 = Node(
        names.SECTION, parent=sampling_description
    )
    sampling_description.add_child(sampling_description_section_2)

    sampling_description_para_1 = Node(names.PARA, parent=sampling_description_section_1)
    sampling_description_section_1.add_child(sampling_description_para_1)
    sampling_description_para_1.content = "This is sampling_description Para 1"

    sampling_description_para_2 = Node(names.PARA, parent=sampling_description_section_2)
    sampling_description_section_2.add_child(sampling_description_para_2)
    sampling_description_para_2.content = "This is sampling_description Para 2"

    quality_control = Node(names.QUALITYCONTROL, parent=methods)
    methods.add_child(quality_control)

    quality_control_description = Node(
        names.DESCRIPTION, parent=quality_control
    )
    quality_control.add_child(quality_control_description)

    quality_control_description_section_1 = Node(
        names.SECTION, parent=quality_control_description
    )
    quality_control_description.add_child(
        quality_control_description_section_1
    )

    quality_control_description_section_2 = Node(
        names.SECTION, parent=quality_control_description
    )
    quality_control_description.add_child(
        quality_control_description_section_2
    )

    quality_control_description_para_1 = Node(
        names.PARA, parent=quality_control_description_section_1
    )
    quality_control_description_section_1.add_child(quality_control_description_para_1)
    quality_control_description_para_1.content = (
        "This is quality_control_description Para 1"
    )

    quality_control_description_para_2 = Node(
        names.PARA, parent=quality_control_description_section_2
    )
    quality_control_description_section_2.add_child(quality_control_description_para_2)
    quality_control_description_para_2.content = (
        "This is quality_control_description Para 2"
    )

    quality_control_instrumentation_1 = Node(
        names.INSTRUMENTATION, parent=quality_control
    )
    quality_control.add_child(quality_control_instrumentation_1)
    quality_control_instrumentation_1.content = (
        "Quality Control Instrumentation string 1"
    )

    quality_control_instrumentation_2 = Node(
        names.INSTRUMENTATION, parent=quality_control
    )
    quality_control.add_child(quality_control_instrumentation_2)
    quality_control_instrumentation_2.content = (
        "Quality Control Instrumentation string 2"
    )

    project = Node(names.PROJECT, parent=dataset)
    dataset.add_child(project)

    project_title = Node(names.TITLE, parent=project)
    project.add_child(project_title)
    project_title.content = "The Title of the Project"

    personnel = Node(names.PERSONNEL, parent=project)
    project.add_child(personnel)

    individualName_personnel = Node(names.INDIVIDUALNAME, parent=personnel)
    personnel.add_child(individualName_personnel)

    givenName_personnel = Node(
        names.GIVENNAME, parent=individualName_personnel
    )
    givenName_personnel.content = "Baxter"
    individualName_personnel.add_child(givenName_personnel)

    surName_personnel = Node(names.SURNAME, parent=individualName_personnel)
    surName_personnel.content = "Duffington"
    individualName_personnel.add_child(surName_personnel)

    project_role = Node(names.ROLE, parent=personnel)
    personnel.add_child(project_role)
    project_role.content = "principalInvestigator"

    datatable = Node(names.DATATABLE, parent=dataset)
    dataset.add_child(datatable)

    alternate_identifier = Node(names.ALTERNATEIDENTIFIER, parent=datatable)
    datatable.add_child(alternate_identifier)
    alternate_identifier.add_attribute("system", "EDI")
    alternate_identifier.content = "https://portal.edirepository.org/nis/dataviewer?packageid=edi.23.1&entityid=ce5438fb9318b1f04d06b8a0276754d6"

    entity_name = Node(names.ENTITYNAME, parent=datatable)
    datatable.add_child(entity_name)
    entity_name.content = "14CO2 data"

    entity_description = Node(names.ENTITYDESCRIPTION, parent=datatable)
    datatable.add_child(entity_description)
    entity_description.content = "Radiocarbon content of soil-respired air"

    physical = Node(names.PHYSICAL, parent=datatable)
    datatable.add_child(physical)
    physical.add_attribute("system", "EDI")

    object_name = Node(names.OBJECTNAME, parent=physical)
    physical.add_child(object_name)
    object_name.content = "tvan_14co2.jk.data.csv"

    size = Node(names.SIZE, parent=physical)
    physical.add_child(size)
    size.add_attribute("unit", "MB")
    size.content = "20"

    authentication = Node(names.AUTHENTICATION, parent=physical)
    physical.add_child(authentication)
    authentication.add_attribute("method", "MD5")
    authentication.content = "7c9e8eb43ab7b52fa9b1db7de756d6a2"

    compression_method = Node(names.COMPRESSIONMETHOD, parent=physical)
    physical.add_child(compression_method)
    compression_method.content = "zip"

    encoding_method = Node(names.ENCODINGMETHOD, parent=physical)
    physical.add_child(encoding_method)
    encoding_method.content = "base64"

    character_encoding = Node(names.CHARACTERENCODING, parent=physical)
    physical.add_child(character_encoding)
    character_encoding.content = "UTF-8"

    data_format = Node(names.DATAFORMAT, parent=physical)
    physical.add_child(data_format)

    text_format = Node(names.TEXTFORMAT, parent=data_format)
    data_format.add_child(text_format)

    num_header_lines = Node(names.NUMHEADERLINES, parent=text_format)
    text_format.add_child(num_header_lines)
    num_header_lines.content = "3"

    num_footer_lines = Node(names.NUMFOOTERLINES, parent=text_format)
    text_format.add_child(num_footer_lines)
    num_footer_lines.content = "0"

    record_delimiter = Node(names.RECORDDELIMITER, parent=text_format)
    text_format.add_child(record_delimiter)
    record_delimiter.content = "\\r\\n"

    physical_line_delimiter = Node(
        names.PHYSICALLINEDELIMITER, parent=text_format
    )
    text_format.add_child(physical_line_delimiter)
    physical_line_delimiter.content = "\\r\\n"

    num_physical_lines_per_record = Node(
        names.NUMPHYSICALLINESPERRECORD, parent=text_format
    )
    text_format.add_child(num_physical_lines_per_record)
    num_physical_lines_per_record.content = "1"

    max_record_length = Node(names.MAXRECORDLENGTH, parent=text_format)
    text_format.add_child(max_record_length)
    max_record_length.content = "597"

    attribute_orientation = Node(
        names.ATTRIBUTEORIENTATION, parent=text_format
    )
    text_format.add_child(attribute_orientation)
    attribute_orientation.content = "column"

    simple_delimited = Node(names.SIMPLEDELIMITED, parent=text_format)
    text_format.add_child(simple_delimited)

    field_delimiter = Node(names.FIELDDELIMITER, parent=simple_delimited)
    simple_delimited.add_child(field_delimiter)
    field_delimiter.content = "\\t"

    collapse_delimiters = Node(
        names.COLLAPSEDELIMITERS, parent=simple_delimited
    )
    simple_delimited.add_child(collapse_delimiters)
    collapse_delimiters.content = "no"

    quote_character = Node(names.QUOTECHARACTER, parent=simple_delimited)
    simple_delimited.add_child(quote_character)
    quote_character.content = "'"

    literal_character = Node(names.LITERALCHARACTER, parent=simple_delimited)
    simple_delimited.add_child(literal_character)
    literal_character.content = (
        "/"  # Usually a backslash, but that causes an error
    )

    complex = Node(names.COMPLEX, parent=text_format)
    # text_format.add_child(complex)

    text_fixed = Node(names.TEXTFIXED, parent=complex)
    complex.add_child(text_fixed)

    field_width = Node(names.FIELDWIDTH, parent=text_fixed)
    text_fixed.add_child(field_width)
    field_width.content = "12"

    line_number = Node(names.LINENUMBER, parent=text_fixed)
    text_fixed.add_child(line_number)
    line_number.content = "3"

    field_start_column = Node(names.FIELDSTARTCOLUMN, parent=text_fixed)
    text_fixed.add_child(field_start_column)
    field_start_column.content = "58"

    text_delimited = Node(names.TEXTDELIMITED, parent=complex)
    complex.add_child(text_delimited)

    field_delimiter_2 = Node(names.FIELDDELIMITER, parent=text_delimited)
    text_delimited.add_child(field_delimiter_2)
    field_delimiter_2.content = ","

    collapse_delimiters_2 = Node(
        names.COLLAPSEDELIMITERS, parent=text_delimited
    )
    text_delimited.add_child(collapse_delimiters_2)
    collapse_delimiters_2.content = "no"

    line_number_2 = Node(names.LINENUMBER, parent=text_delimited)
    text_delimited.add_child(line_number_2)
    line_number_2.content = "2"

    quote_character_2 = Node(names.QUOTECHARACTER, parent=text_delimited)
    text_delimited.add_child(quote_character_2)
    quote_character_2.content = "'"

    literal_character_2 = Node(names.LITERALCHARACTER, parent=text_delimited)
    text_delimited.add_child(literal_character_2)
    literal_character_2.content = (
        "/"  # Usually a backslash, but that causes an error
    )

    externally_defined_format = Node(
        names.EXTERNALLYDEFINEDFORMAT, parent=data_format
    )

    format_name = Node(names.FORMATNAME, parent=externally_defined_format)
    externally_defined_format.add_child(format_name)
    format_name.content = "Microsoft Excel"

    format_version = Node(
        names.FORMATVERSION, parent=externally_defined_format
    )
    externally_defined_format.add_child(format_version)
    format_version.content = "2000 (9.0.2720)"

    binary_raster_format = Node(names.BINARYRASTERFORMAT, parent=data_format)

    row_column_orientation = Node(
        names.ROWCOLUMNORIENTATION, parent=binary_raster_format
    )
    binary_raster_format.add_child(row_column_orientation)
    row_column_orientation.content = "column"

    multi_band = Node(names.MULTIBAND, parent=binary_raster_format)
    binary_raster_format.add_child(multi_band)

    nbands = Node(names.NBANDS, parent=multi_band)
    multi_band.add_child(nbands)
    nbands.content = "2"

    layout = Node(names.LAYOUT, parent=multi_band)
    multi_band.add_child(layout)
    layout.content = "bil"

    nbits = Node(names.NBITS, parent=binary_raster_format)
    binary_raster_format.add_child(nbits)
    nbits.content = "16"

    byteorder = Node(names.BYTEORDER, parent=binary_raster_format)
    binary_raster_format.add_child(byteorder)
    byteorder.content = "little-endian"

    skipbytes = Node(names.SKIPBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(skipbytes)
    skipbytes.content = "0"

    bandrowbytes = Node(names.BANDROWBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(bandrowbytes)
    bandrowbytes.content = "3"

    totalrowbytes = Node(names.TOTALROWBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(totalrowbytes)
    totalrowbytes.content = "8"

    bandgapbytes = Node(names.BANDGAPBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(bandgapbytes)
    bandgapbytes.content = "1"

    distribution = Node(names.DISTRIBUTION, parent=physical)
    physical.add_child(distribution)

    online = Node(names.ONLINE, parent=distribution)
    distribution.add_child(online)

    online_description = Node(names.ONLINEDESCRIPTION, parent=online)
    online.add_child(online_description)
    online_description.content = "network server"

    url = Node(names.URL, parent=online)
    online.add_child(url)
    url.add_attribute("function", "download")
    url.content = "https://data.abc.edu/somedataset/data"

    # connection = Node(names.CONNECTION, parent=online)
    # online.add_child(connection)
    # connection.content = (
    #     "This is not a real connection element because it is only a string."
    # )

    offline = Node(names.OFFLINE, parent=distribution)

    medium_name = Node(names.MEDIUMNAME, parent=offline)
    offline.add_child(medium_name)
    medium_name.content = "hardcopy"

    medium_density = Node(names.MEDIUMDENSITY, parent=offline)
    offline.add_child(medium_density)
    medium_density.content = "High Density (HD)"

    medium_density_units = Node(names.MEDIUMDENSITYUNITS, parent=offline)
    offline.add_child(medium_density_units)
    medium_density_units.content = "B/cm"

    medium_volume = Node(names.MEDIUMVOLUME, parent=offline)
    offline.add_child(medium_volume)
    medium_volume.content = "650 MB"

    medium_format = Node(names.MEDIUMFORMAT, parent=offline)
    offline.add_child(medium_format)
    medium_format.content = "NTFS"

    medium_note = Node(names.MEDIUMNOTE, parent=offline)
    offline.add_child(medium_note)
    medium_note.content = (
        "Some additional pertinent information about the media"
    )

    inline = Node(names.INLINE, parent=distribution)
    inline.content = "1,2,3,4,5,6,7,8,9"

    datatable_access = Node(names.ACCESS, parent=distribution)
    datatable_access.add_attribute("authSystem", "pasta")
    datatable_access.add_attribute("order", "allowFirst")
    distribution.add_child(datatable_access)

    datatable_allow = Node(names.ALLOW, parent=datatable_access)
    datatable_access.add_child(datatable_allow)

    datatable_principal = Node(names.PRINCIPAL, parent=datatable_allow)
    datatable_principal.content = "uid=ucarroll,o=EDI,dc=edirepository,dc=org"
    datatable_allow.add_child(datatable_principal)

    datatable_permission = Node(names.PERMISSION, parent=datatable_allow)
    datatable_permission.content = "write"
    datatable_allow.add_child(datatable_permission)

    attribute_list = Node(names.ATTRIBUTELIST, parent=datatable)
    datatable.add_child(attribute_list)

    attribute_1 = Node(names.ATTRIBUTE, parent=attribute_list)
    attribute_list.add_child(attribute_1)
    attribute_name_1 = Node(names.ATTRIBUTENAME, parent=attribute_1)
    attribute_1.add_child(attribute_name_1)
    attribute_name_1.content = "att-1"
    att_label_1 = Node(names.ATTRIBUTELABEL, parent=attribute_1)
    attribute_1.add_child(att_label_1)
    att_label_1.content = "Attribute Label 1"
    att_def_1 = Node(names.ATTRIBUTEDEFINITION)
    att_def_1.content = "Attribute 1 definition"
    attribute_1.add_child(att_def_1)
    ms_1 = Node(names.MEASUREMENTSCALE, parent=attribute_1)
    attribute_1.add_child(ms_1)
    nominal = Node(names.NOMINAL, parent=ms_1)
    ms_1.add_child(nominal)
    non_numeric_domain_1 = Node(names.NONNUMERICDOMAIN, parent=nominal)
    nominal.add_child(non_numeric_domain_1)
    enumerated_domain = Node(
        names.ENUMERATEDDOMAIN, parent=non_numeric_domain_1
    )
    non_numeric_domain_1.add_child(enumerated_domain)
    enumerated_domain.add_attribute("enforced", "yes")
    code_definition = Node(names.CODEDEFINITION, parent=enumerated_domain)
    enumerated_domain.add_child(code_definition)
    code_definition.add_attribute("order", "3")
    code = Node(names.CODE, parent=code_definition)
    code_definition.add_child(code)
    code.content = "HIGH"
    definition = Node(names.DEFINITION, parent=code_definition)
    code_definition.add_child(definition)
    definition.content = "high density, above 10 per square meter"
    source = Node(names.SOURCE, parent=code_definition)
    code_definition.add_child(source)
    source.content = "ISO country codes"
    external_code_set = Node(names.EXTERNALCODESET, parent=enumerated_domain)
    # enumerated_domain.add_child(external_code_set)
    codeset_name = Node(names.CODESETNAME, parent=external_code_set)
    external_code_set.add_child(codeset_name)
    codeset_name.content = "FIPS State Abbreviation Codes"
    citation = Node(names.CITATION, parent=external_code_set)
    external_code_set.add_child(citation)
    citation.content = "This is not an actual citation element"
    codeset_url = Node(names.CODESETURL, parent=external_code_set)
    external_code_set.add_child(codeset_url)
    codeset_url.content = (
        "https://codesets.abc.edu/fips_state_abbreviation_codes"
    )
    entity_code_list = Node(names.ENTITYCODELIST, parent=enumerated_domain)
    # enumerated_domain.add_child(entity_code_list)
    entity_reference = Node(names.ENTITYREFERENCE, parent=entity_code_list)
    entity_code_list.add_child(entity_reference)
    entity_reference.content = "entity_id_1"
    value_attribute_reference = Node(
        names.VALUEATTRIBUTEREFERENCE, parent=entity_code_list
    )
    entity_code_list.add_child(value_attribute_reference)
    value_attribute_reference.content = "attribute_id_1"
    definition_attribute_reference = Node(
        names.DEFINITIONATTRIBUTEREFERENCE, parent=entity_code_list
    )
    entity_code_list.add_child(definition_attribute_reference)
    definition_attribute_reference.content = "attribute_id_1"
    order_attribute_reference = Node(
        names.ORDERATTRIBUTEREFERENCE, parent=entity_code_list
    )
    entity_code_list.add_child(order_attribute_reference)
    order_attribute_reference.content = "attribute_id_1"

    attribute_2 = Node(names.ATTRIBUTE, parent=attribute_list)
    attribute_list.add_child(attribute_2)
    attribute_name_2 = Node(names.ATTRIBUTENAME, parent=attribute_2)
    attribute_2.add_child(attribute_name_2)
    attribute_name_2.content = "att-2"
    att_label_2 = Node(names.ATTRIBUTELABEL, parent=attribute_2)
    attribute_2.add_child(att_label_2)
    att_label_2.content = "Attribute Label 2"
    att_def_2 = Node(names.ATTRIBUTEDEFINITION)
    att_def_2.content = "Attribute 2 definition"
    attribute_2.add_child(att_def_2)
    ms_2 = Node(names.MEASUREMENTSCALE, parent=attribute_2)
    attribute_2.add_child(ms_2)
    ordinal = Node(names.ORDINAL, parent=ms_2)
    ms_2.add_child(ordinal)
    non_numeric_domain_2 = Node(names.NONNUMERICDOMAIN, parent=ordinal)
    ordinal.add_child(non_numeric_domain_2)
    text_domain = Node(names.TEXTDOMAIN, parent=non_numeric_domain_2)
    non_numeric_domain_2.add_child(text_domain)
    definition_2 = Node(names.DEFINITION, parent=text_domain)
    text_domain.add_child(definition_2)
    definition_2.content = "US telephone numbers in the format (999) 888-7777"
    pattern = Node(names.PATTERN, parent=text_domain)
    text_domain.add_child(pattern)
    pattern.content = "[0-9a-zA-Z] matches simple alphanumeric strings"
    source_2 = Node(names.SOURCE, parent=text_domain)
    text_domain.add_child(source_2)
    source_2.content = "ISO country codes"

    attribute_3 = Node(names.ATTRIBUTE, parent=attribute_list)
    attribute_list.add_child(attribute_3)
    attribute_name_3 = Node(names.ATTRIBUTENAME, parent=attribute_3)
    attribute_3.add_child(attribute_name_3)
    attribute_name_3.content = "att-3"
    att_label_3 = Node(names.ATTRIBUTELABEL, parent=attribute_3)
    attribute_3.add_child(att_label_3)
    att_label_3.content = "Attribute Label 3"
    att_def_3 = Node(names.ATTRIBUTEDEFINITION)
    att_def_3.content = "Attribute 3 definition"
    attribute_3.add_child(att_def_3)
    storage_type = Node(names.STORAGETYPE, parent=attribute_3)
    attribute_3.add_child(storage_type)
    storage_type.add_attribute(
        "typeSystem", "http://www.w3.org/2001/XMLSchema-datatypes"
    )
    storage_type.content = "integer"
    ms_3 = Node(names.MEASUREMENTSCALE, parent=attribute_3)
    attribute_3.add_child(ms_3)
    interval = Node(names.INTERVAL, parent=ms_3)
    ms_3.add_child(interval)
    unit = Node(names.UNIT, parent=interval)
    interval.add_child(unit)
    standard_unit = Node(names.STANDARDUNIT, parent=unit)
    unit.add_child(standard_unit)
    standard_unit.content = "meter"
    custom_unit = Node(names.CUSTOMUNIT, parent=unit)
    # unit.add_child(custom_unit)
    custom_unit.content = "metersPerOneThirdGram"
    precision = Node(names.PRECISION, parent=interval)
    interval.add_child(precision)
    precision.content = "0.1"
    numeric_domain = Node(names.NUMERICDOMAIN, parent=interval)
    interval.add_child(numeric_domain)
    number_type = Node(names.NUMBERTYPE, parent=numeric_domain)
    numeric_domain.add_child(number_type)
    number_type.content = "integer"
    bounds = Node(names.BOUNDS, parent=numeric_domain)
    numeric_domain.add_child(bounds)
    minimum = Node(names.MINIMUM, parent=bounds)
    bounds.add_child(minimum)
    minimum.add_attribute("exclusive", "false")
    minimum.content = "0"
    maximum = Node(names.MAXIMUM, parent=bounds)
    bounds.add_child(maximum)
    maximum.add_attribute("exclusive", "false")
    maximum.content = "10"

    attribute_4 = Node(names.ATTRIBUTE, parent=attribute_list)
    attribute_list.add_child(attribute_4)
    attribute_name_4 = Node(names.ATTRIBUTENAME, parent=attribute_4)
    attribute_4.add_child(attribute_name_4)
    attribute_name_4.content = "att-4"
    att_label_4 = Node(names.ATTRIBUTELABEL, parent=attribute_4)
    attribute_4.add_child(att_label_4)
    att_label_4.content = "Attribute Label 4"
    att_def_4 = Node(names.ATTRIBUTEDEFINITION)
    att_def_4.content = "Attribute 4 definition"
    attribute_4.add_child(att_def_4)
    ms_4 = Node(names.MEASUREMENTSCALE, parent=attribute_4)
    attribute_4.add_child(ms_4)
    ratio = Node(names.RATIO, parent=ms_4)
    ms_4.add_child(ratio)
    unit_ratio = Node(names.UNIT, parent=ratio)
    ratio.add_child(unit_ratio)
    standard_unit_ratio = Node(names.STANDARDUNIT, parent=unit_ratio)
    # unit_ratio.add_child(standard_unit_ratio)
    standard_unit_ratio.content = "grams"
    custom_unit_ratio = Node(names.CUSTOMUNIT, parent=unit_ratio)
    unit_ratio.add_child(custom_unit_ratio)
    custom_unit_ratio.content = "gramsPerOneThirdMeter"
    precision_ratio = Node(names.PRECISION, parent=ratio)
    ratio.add_child(precision_ratio)
    precision_ratio.content = "0.01"
    numeric_domain_ratio = Node(names.NUMERICDOMAIN, parent=ratio)
    ratio.add_child(numeric_domain_ratio)
    number_type_ratio = Node(names.NUMBERTYPE, parent=numeric_domain_ratio)
    numeric_domain_ratio.add_child(number_type_ratio)
    number_type_ratio.content = "real"
    bounds_ratio = Node(names.BOUNDS, parent=numeric_domain_ratio)
    numeric_domain_ratio.add_child(bounds_ratio)
    minimum_ratio = Node(names.MINIMUM, parent=bounds_ratio)
    bounds_ratio.add_child(minimum_ratio)
    minimum_ratio.add_attribute("exclusive", "false")
    minimum_ratio.content = "0.0"
    maximum_ratio = Node(names.MAXIMUM, parent=bounds_ratio)
    bounds_ratio.add_child(maximum_ratio)
    maximum_ratio.add_attribute("exclusive", "false")
    maximum_ratio.content = "10.0"

    attribute_5 = Node(names.ATTRIBUTE, parent=attribute_list)
    attribute_list.add_child(attribute_5)
    attribute_name_5 = Node(names.ATTRIBUTENAME, parent=attribute_5)
    attribute_5.add_child(attribute_name_5)
    attribute_name_5.content = "att-5"
    att_label_5 = Node(names.ATTRIBUTELABEL, parent=attribute_5)
    attribute_5.add_child(att_label_5)
    att_label_5.content = "Attribute Label 5"
    att_def_5 = Node(names.ATTRIBUTEDEFINITION)
    att_def_5.content = "Attribute 5 definition"
    attribute_5.add_child(att_def_5)
    ms_5 = Node(names.MEASUREMENTSCALE, parent=attribute_5)
    attribute_5.add_child(ms_5)
    datetime = Node(names.DATETIME, parent=ms_5)
    ms_5.add_child(datetime)
    format_string = Node(names.FORMATSTRING, parent=datetime)
    datetime.add_child(format_string)
    format_string.content = "YYYY-MM-DD"
    datetime_precision = Node(names.DATETIMEPRECISION, parent=datetime)
    datetime.add_child(datetime_precision)
    datetime_precision.content = "1"
    datetime_domain = Node(names.DATETIMEDOMAIN, parent=datetime)
    datetime.add_child(datetime_domain)
    datetime_domain.add_attribute("id", "dtd-1")
    bounds_datetime = Node(names.BOUNDS, parent=datetime_domain)
    datetime_domain.add_child(bounds_datetime)
    minimum_datetime = Node(names.MINIMUM, parent=bounds_datetime)
    bounds_datetime.add_child(minimum_datetime)
    minimum_datetime.add_attribute("exclusive", "false")
    minimum_datetime.content = "2003-10-15"
    maximum_datetime = Node(names.MAXIMUM, parent=bounds_datetime)
    bounds_datetime.add_child(maximum_datetime)
    maximum_datetime.add_attribute("exclusive", "false")
    maximum_datetime.content = "2019-06-30"

    missing_value_code = Node(names.MISSINGVALUECODE, parent=attribute_1)
    attribute_1.add_child(missing_value_code)
    code_mvc = Node(names.CODE, parent=missing_value_code)
    missing_value_code.add_child(code_mvc)
    code_mvc.content = "SDT"
    code_explanation = Node(names.CODEEXPLANATION, parent=missing_value_code)
    missing_value_code.add_child(code_explanation)
    code_explanation.content = "Sensor Downtime"
    accuracy = Node(names.ACCURACY, parent=attribute_1)
    attribute_1.add_child(accuracy)
    attribute_accuracy_report = Node(
        names.ATTRIBUTEACCURACYREPORT, parent=accuracy
    )
    accuracy.add_child(attribute_accuracy_report)
    attribute_accuracy_report.content = "An explanation of the accuracy of the observation recorded in this attribute"
    qaaa = Node(names.QUANTITATIVEATTRIBUTEACCURACYASSESSMENT, parent=accuracy)
    accuracy.add_child(qaaa)
    attribute_accuracy_value = Node(names.ATTRIBUTEACCURACYVALUE, parent=qaaa)
    qaaa.add_child(attribute_accuracy_value)
    attribute_accuracy_value.content = (
        "An estimate of the accuracy of the identification"
    )
    attribute_accuracy_explanation = Node(
        names.ATTRIBUTEACCURACYEXPLANATION, parent=qaaa
    )
    qaaa.add_child(attribute_accuracy_explanation)
    attribute_accuracy_explanation.content = 'The identification of the test that yielded the "Attribute Accuracy Value"'

    case_sensitive = Node(names.CASESENSITIVE, parent=datatable)
    datatable.add_child(case_sensitive)
    case_sensitive.content = "yes"

    number_of_records = Node(names.NUMBEROFRECORDS, parent=datatable)
    datatable.add_child(number_of_records)
    number_of_records.content = "454"

    other_entity = Node(names.OTHERENTITY, parent=dataset)
    dataset.add_child(other_entity)

    alternate_identifier = Node(names.ALTERNATEIDENTIFIER, parent=other_entity)
    other_entity.add_child(alternate_identifier)
    alternate_identifier.add_attribute("system", "EDI")
    alternate_identifier.content = "https://portal.edirepository.org/nis/dataviewer?packageid=edi.23.1&entityid=ce5438fb9318b1f04d06b8a0276754d6"

    entity_name = Node(names.ENTITYNAME, parent=other_entity)
    other_entity.add_child(entity_name)
    entity_name.content = "Some Other Entity Data"

    entity_description = Node(names.ENTITYDESCRIPTION, parent=other_entity)
    other_entity.add_child(entity_description)
    entity_description.content = "Radiocarbon content of soil-respired air"

    physical = Node(names.PHYSICAL, parent=other_entity)
    other_entity.add_child(physical)
    physical.add_attribute("system", "EDI")

    object_name = Node(names.OBJECTNAME, parent=physical)
    physical.add_child(object_name)
    object_name.content = "tvan_14co2.jk.data.csv"

    size = Node(names.SIZE, parent=physical)
    physical.add_child(size)
    size.add_attribute("unit", "MB")
    size.content = "20"

    authentication = Node(names.AUTHENTICATION, parent=physical)
    physical.add_child(authentication)
    authentication.add_attribute("method", "password")
    authentication.content = "LDAP"

    compression_method = Node(names.COMPRESSIONMETHOD, parent=physical)
    physical.add_child(compression_method)
    compression_method.content = "zip"

    encoding_method = Node(names.ENCODINGMETHOD, parent=physical)
    physical.add_child(encoding_method)
    encoding_method.content = "base64"

    character_encoding = Node(names.CHARACTERENCODING, parent=physical)
    physical.add_child(character_encoding)
    character_encoding.content = "UTF-8"

    data_format = Node(names.DATAFORMAT, parent=physical)
    physical.add_child(data_format)

    text_format = Node(names.TEXTFORMAT, parent=data_format)
    data_format.add_child(text_format)

    num_header_lines = Node(names.NUMHEADERLINES, parent=text_format)
    text_format.add_child(num_header_lines)
    num_header_lines.content = "3"

    num_footer_lines = Node(names.NUMFOOTERLINES, parent=text_format)
    text_format.add_child(num_footer_lines)
    num_footer_lines.content = "0"

    record_delimiter = Node(names.RECORDDELIMITER, parent=text_format)
    text_format.add_child(record_delimiter)
    record_delimiter.content = "\\r\\n"

    physical_line_delimiter = Node(
        names.PHYSICALLINEDELIMITER, parent=text_format
    )
    text_format.add_child(physical_line_delimiter)
    physical_line_delimiter.content = "\\r\\n"

    num_physical_lines_per_record = Node(
        names.NUMPHYSICALLINESPERRECORD, parent=text_format
    )
    text_format.add_child(num_physical_lines_per_record)
    num_physical_lines_per_record.content = "1"

    max_record_length = Node(names.MAXRECORDLENGTH, parent=text_format)
    text_format.add_child(max_record_length)
    max_record_length.content = "597"

    attribute_orientation = Node(
        names.ATTRIBUTEORIENTATION, parent=text_format
    )
    text_format.add_child(attribute_orientation)
    attribute_orientation.content = "column"

    simple_delimited = Node(names.SIMPLEDELIMITED, parent=text_format)
    # text_format.add_child(simple_delimited)

    field_delimiter = Node(names.FIELDDELIMITER, parent=simple_delimited)
    simple_delimited.add_child(field_delimiter)
    field_delimiter.content = "\\t"

    collapse_delimiters = Node(
        names.COLLAPSEDELIMITERS, parent=simple_delimited
    )
    simple_delimited.add_child(collapse_delimiters)
    collapse_delimiters.content = "no"

    quote_character = Node(names.QUOTECHARACTER, parent=simple_delimited)
    simple_delimited.add_child(quote_character)
    quote_character.content = "'"

    literal_character = Node(names.LITERALCHARACTER, parent=simple_delimited)
    simple_delimited.add_child(literal_character)
    literal_character.content = (
        "/"  # Usually a backslash, but that causes an error
    )

    complex = Node(names.COMPLEX, parent=text_format)
    text_format.add_child(complex)

    text_fixed = Node(names.TEXTFIXED, parent=complex)
    complex.add_child(text_fixed)

    field_width = Node(names.FIELDWIDTH, parent=text_fixed)
    text_fixed.add_child(field_width)
    field_width.content = "12"

    line_number = Node(names.LINENUMBER, parent=text_fixed)
    text_fixed.add_child(line_number)
    line_number.content = "3"

    field_start_column = Node(names.FIELDSTARTCOLUMN, parent=text_fixed)
    text_fixed.add_child(field_start_column)
    field_start_column.content = "58"

    text_delimited = Node(names.TEXTDELIMITED, parent=complex)
    complex.add_child(text_delimited)

    field_delimiter_2 = Node(names.FIELDDELIMITER, parent=text_delimited)
    text_delimited.add_child(field_delimiter_2)
    field_delimiter_2.content = ","

    collapse_delimiters_2 = Node(
        names.COLLAPSEDELIMITERS, parent=text_delimited
    )
    text_delimited.add_child(collapse_delimiters_2)
    collapse_delimiters_2.content = "no"

    line_number_2 = Node(names.LINENUMBER, parent=text_delimited)
    text_delimited.add_child(line_number_2)
    line_number_2.content = "2"

    quote_character_2 = Node(names.QUOTECHARACTER, parent=text_delimited)
    text_delimited.add_child(quote_character_2)
    quote_character_2.content = "'"

    literal_character_2 = Node(names.LITERALCHARACTER, parent=text_delimited)
    text_delimited.add_child(literal_character_2)
    literal_character_2.content = (
        "/"  # Usually a backslash, but that causes an error
    )

    externally_defined_format = Node(
        names.EXTERNALLYDEFINEDFORMAT, parent=data_format
    )
    # data_format.add_child(externally_defined_format)

    format_name = Node(names.FORMATNAME, parent=externally_defined_format)
    externally_defined_format.add_child(format_name)
    format_name.content = "Microsoft Excel"

    format_version = Node(
        names.FORMATVERSION, parent=externally_defined_format
    )
    externally_defined_format.add_child(format_version)
    format_version.content = "2000 (9.0.2720)"

    binary_raster_format = Node(names.BINARYRASTERFORMAT, parent=data_format)
    # data_format.add_child(binary_raster_format)

    row_column_orientation = Node(
        names.ROWCOLUMNORIENTATION, parent=binary_raster_format
    )
    binary_raster_format.add_child(row_column_orientation)
    row_column_orientation.content = "column"

    multi_band = Node(names.MULTIBAND, parent=binary_raster_format)
    binary_raster_format.add_child(multi_band)

    nbands = Node(names.NBANDS, parent=multi_band)
    multi_band.add_child(nbands)
    nbands.content = "2"

    layout = Node(names.LAYOUT, parent=multi_band)
    multi_band.add_child(layout)
    layout.content = "bil"

    nbits = Node(names.NBITS, parent=binary_raster_format)
    binary_raster_format.add_child(nbits)
    nbits.content = "16"

    byteorder = Node(names.BYTEORDER, parent=binary_raster_format)
    binary_raster_format.add_child(byteorder)
    byteorder.content = "little-endian"

    skipbytes = Node(names.SKIPBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(skipbytes)
    skipbytes.content = "0"

    bandrowbytes = Node(names.BANDROWBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(bandrowbytes)
    bandrowbytes.content = "3"

    totalrowbytes = Node(names.TOTALROWBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(totalrowbytes)
    totalrowbytes.content = "8"

    bandgapbytes = Node(names.BANDGAPBYTES, parent=binary_raster_format)
    binary_raster_format.add_child(bandgapbytes)
    bandgapbytes.content = "1"

    distribution = Node(names.DISTRIBUTION, parent=physical)
    physical.add_child(distribution)

    online = Node(names.ONLINE, parent=distribution)
    # distribution.add_child(online)

    online_description = Node(names.ONLINEDESCRIPTION, parent=online)
    online.add_child(online_description)
    online_description.content = "network server"

    url = Node(names.URL, parent=online)
    online.add_child(url)
    url.add_attribute("function", "download")
    url.content = "https://data.abc.edu/somedataset/other_entity_data"

    # connection = Node(names.CONNECTION, parent=online)
    # online.add_child(connection)
    # connection.content = (
    #     "This is not a real connection element because it is only a string."
    # )

    offline = Node(names.OFFLINE, parent=distribution)
    distribution.add_child(offline)

    medium_name = Node(names.MEDIUMNAME, parent=offline)
    offline.add_child(medium_name)
    medium_name.content = "hardcopy"

    medium_density = Node(names.MEDIUMDENSITY, parent=offline)
    offline.add_child(medium_density)
    medium_density.content = "High Density (HD)"

    medium_density_units = Node(names.MEDIUMDENSITYUNITS, parent=offline)
    offline.add_child(medium_density_units)
    medium_density_units.content = "B/cm"

    medium_volume = Node(names.MEDIUMVOLUME, parent=offline)
    offline.add_child(medium_volume)
    medium_volume.content = "650 MB"

    medium_format = Node(names.MEDIUMFORMAT, parent=offline)
    offline.add_child(medium_format)
    medium_format.content = "NTFS"

    medium_note = Node(names.MEDIUMNOTE, parent=offline)
    offline.add_child(medium_note)
    medium_note.content = (
        "Some additional pertinent information about the media"
    )

    inline = Node(names.INLINE, parent=distribution)
    # distribution.add_child(inline)
    inline.content = "1,2,3,4,5,6,7,8,9"

    other_entity_access = Node(names.ACCESS, parent=distribution)
    other_entity_access.add_attribute("authSystem", "pasta")
    other_entity_access.add_attribute("order", "allowFirst")
    distribution.add_child(other_entity_access)

    other_entity_allow = Node(names.ALLOW, parent=other_entity_access)
    other_entity_access.add_child(other_entity_allow)

    other_entity_principal = Node(names.PRINCIPAL, parent=other_entity_allow)
    other_entity_principal.content = (
        "uid=ucarroll,o=EDI,dc=edirepository,dc=org"
    )
    other_entity_allow.add_child(other_entity_principal)

    other_entity_permission = Node(names.PERMISSION, parent=other_entity_allow)
    other_entity_permission.content = "write"
    other_entity_allow.add_child(other_entity_permission)

    entity_type = Node(names.ENTITYTYPE, parent=other_entity)
    other_entity.add_child(entity_type)
    entity_type.content = "Unknown Entity Type"

    datatable_rule = rule.get_rule(names.DATATABLE)
    try:
        datatable_rule.validate_rule(datatable)
    except MetapypeRuleError as e:
        logger.error(e)

    text_format_rule = rule.get_rule(names.TEXTFORMAT)
    try:
        text_format_rule.validate_rule(text_format)
    except MetapypeRuleError as e:
        logger.error(e)

    online_rule = rule.get_rule(names.ONLINE)
    try:
        online_rule.validate_rule(online)
    except MetapypeRuleError as e:
        logger.error(e)

    node = Node.get_node_instance(access.id)

    access_rule = rule.get_rule(names.ACCESS)

    try:
        access_rule.validate_rule(access)
    except MetapypeRuleError as e:
        logger.error(e)
    # print(access_rule.attributes)
    # print(access_rule.children)
    # print(access_rule.content_rules)

    try:
        validate.node(access)
    except MetapypeRuleError as e:
        logger.error(e)

    try:
        errors = []
        validate.tree(eml, errors)
    except MetapypeRuleError as e:
        logger.error(e)

    for error in errors:
        print(error)

    # print(evaluate.node(title))
    #
    # mp_io.graph(eml, 0)
    #
    # attr = access_rule.attributes
    # print(attr)
    # children = access_rule.children
    # print(children)
    #
    # print(access.list_attributes())

    json_str = metapype_io.to_json(eml)
    # print(json_str)
    with open("test_eml.json", "w") as f:
        f.write(json_str)

    node = metapype_io.from_json(json_str)

    try:
        validate.tree(node)
    except MetapypeRuleError as e:
        logger.error(e)
    #
    xml = export.to_xml(eml)
    metapype_io.from_xml(xml)
    # print(xml)
    with open("test_eml.xml", "w") as f:
        f.write(xml)

    # warnings = []
    # evaluate.tree(eml, warnings)
    # for w in warnings:
    #     print(f"{w}")
    #
    # other_entity_rule = rule.get_rule(names.OTHERENTITY)
    # try:
    #     other_entity_rule.validate_rule(other_entity)
    # except MetapypeRuleError as e:
    #     logger.error(e)

    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: __init__

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    1/16/19
"""
import importlib.resources


version = importlib.resources.read_text("metapype", "VERSION.txt")
__version__ = version.strip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: evaluate

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    6/21/18
"""
import daiquiri

from metapype.eml import names
from metapype.model.node import Node
from metapype.eml.evaluation_warnings import EvaluationWarning


logger = daiquiri.getLogger("evaluate: " + __name__)


def get_text_content(text_node: Node) -> str:
    # Collect the text under a TextType node.
    # This is intended for use in checking if text is present and that it has enough words, if there is
    #  such a requirement. I.e., it doesn't worry about getting the text in the correct order if, for example, there
    #  are para and markdown nodes interleaved.
    content = text_node.content if text_node.content else ''
    paras = []
    text_node.find_all_descendants(names.PARA, paras)
    for para in paras:
        content += '\n' + para.content
    markdowns = []
    text_node.find_all_descendants(names.MARKDOWN, markdowns)
    for markdown in markdowns:
        content += '\n' + markdown.content
    return content


# ==================== Begin of rules section ====================

def _associated_responsible_party_rule(node: Node) -> list:
    return _responsible_party_rule(node)


def _contact_rule(node: Node) -> list:
    return _responsible_party_rule(node)


def _creator_rule(node: Node) -> list:
    return _responsible_party_rule(node)


def _dataset_rule(node: Node) -> list:
    evaluation = []
    abstract_node = None
    coverage_node = None
    datatable_node = None
    intellectual_rights_node = None
    keywordset_nodes = []
    methods_node = None
    project_node = None
    for child in node.children:
        if child.name == names.ABSTRACT:
            abstract_node = child
        elif child.name == names.COVERAGE:
            coverage_node = child
        elif child.name == names.DATATABLE:
            datatable_node = child
        elif child.name == names.INTELLECTUALRIGHTS:
            intellectual_rights_node = child
        elif child.name == names.KEYWORDSET:
            keywordset_nodes.append(child)
        elif child.name == names.METHODS:
            methods_node = child
        elif child.name == names.PROJECT:
            project_node = child

    if abstract_node:
        content = get_text_content(abstract_node)
        if content:
            words = content.split()
            if len(words) < 20:
                evaluation.append((
                    EvaluationWarning.DATASET_ABSTRACT_TOO_SHORT,
                    f'Consider increasing the length of the dataset''s abstract.',
                    node
                ))
        else:
            evaluation.append((
                EvaluationWarning.DATASET_ABSTRACT_MISSING,
                f'A dataset abstract should be provided.',
                node
            ))
    else:
        evaluation.append((
            EvaluationWarning.DATASET_ABSTRACT_MISSING,
            f'A dataset abstract should be provided.',
            node
        ))
    if not (coverage_node and coverage_node.children):
        evaluation.append((
            EvaluationWarning.DATASET_COVERAGE_MISSING,
            f'At least one coverage element should be present in a dataset. I.e., at least one of Geographic Coverage,'
            f' Temporal Coverage, or Taxonomic Coverage should be specified.',
            node
        ))
    if not datatable_node:
        evaluation.append((
            EvaluationWarning.DATATABLE_MISSING,
            f'A dataset should contain at least one Data Table.',
            node
        ))
    if not (intellectual_rights_node and intellectual_rights_node.content):
        evaluation.append((
            EvaluationWarning.INTELLECTUAL_RIGHTS_MISSING,
            f'An Intellectual Rights policy should be specified.',
            node
        ))
    if not keywordset_nodes:
        evaluation.append((
            EvaluationWarning.KEYWORDS_MISSING,
            f'Keywords should be provided to make the dataset more discoverable.',
            node
        ))
    else:
        num_keywords = 0
        for keywordset_node in keywordset_nodes:
            keyword_nodes = keywordset_node.find_all_children(names.KEYWORD)
            num_keywords += len(keyword_nodes)
        if num_keywords < 5:
            evaluation.append((
                EvaluationWarning.KEYWORDS_INSUFFICIENT,
                f'Consider adding more keywords to make the dataset more discoverable.',
                node
            ))
    if not methods_node:
        evaluation.append((
            EvaluationWarning.DATASET_METHOD_STEPS_MISSING,
            f'A dataset should contain at least one Method Step.',
            node
        ))
    if not project_node:
        evaluation.append((
            EvaluationWarning.DATASET_PROJECT_MISSING,
            f'A dataset should contain a Project.',
            node
        ))

    return evaluation


def _datatable_rule(node: Node) -> list:
    evaluation = []
    description = any(
        child.name == names.ENTITYDESCRIPTION and child.content
        for child in node.children
    )
    if not description:
        evaluation.append((
            EvaluationWarning.DATATABLE_DESCRIPTION_MISSING,
            f'A data table Description is highly recommended.',
            node
        ))
    return evaluation


def _description_rule(node: Node) -> list:
    # Various description nodes are required but since they have TextType, the rules allow them to be empty.
    # Require them to have nonempty content (including para and markdown children).
    evaluation = []
    warning = None
    content = get_text_content(node)
    if not content:
        parent = node.parent.name
        if parent == 'connectionDefinition':
            warning = EvaluationWarning.CONNECTION_DEFINITION_DESCRIPTION_MISSING
        elif parent == 'designDescription':
            warning = EvaluationWarning.DESIGN_DESCRIPTION_DESCRIPTION_MISSING
        elif parent == 'maintenance':
            warning = EvaluationWarning.MAINTENANCE_DESCRIPTION_MISSING
        elif parent == 'methodStep':
            warning = EvaluationWarning.METHOD_STEP_DESCRIPTION_MISSING
        elif parent == 'procedureStep':
            warning = EvaluationWarning.PROCEDURE_STEP_DESCRIPTION_MISSING
        elif parent == 'qualityControl':
            warning = EvaluationWarning.QUALITY_CONTROL_DESCRIPTION_MISSING
        elif parent == 'samplingDescription':
            warning = EvaluationWarning.SAMPLING_DESCRIPTION_DESCRIPTION_MISSING
        elif parent == 'studyExtent':
            warning = EvaluationWarning.STUDY_EXTENT_DESCRIPTION_MISSING
    if warning:
        evaluation.append((
            warning,
            f'A Description is required.',
            node
        ))
    return evaluation


def _individual_name_rule(node: Node) -> list:
    evaluation = []
    givename = False
    surname = False
    for child in node.children:
        if child.name == names.GIVENNAME and child.content:
            givename = True
        if child.name == names.SURNAME and child.content:
            surname = True
    if givename and surname:
        evaluation = None
    else:
        evaluation.append((
            EvaluationWarning.INDIVIDUAL_NAME_INCOMPLETE,
            f'An individual\'s name should have both a "{names.GIVENNAME}" and a "{names.SURNAME}"',
            node
        ))
    return evaluation


def _responsible_party_rule(node: Node) -> list:
    evaluation = []
    userid = False
    orcid = False
    email = False
    for child in node.children:
        if child.name == names.USERID and child.content:
            userid = True
            if child.attributes.get('directory') == 'https://orcid.org':
                orcid = True
        if child.name == names.ELECTRONICMAILADDRESS and child.content:
            email = True
    if not orcid:
        evaluation.append((
            EvaluationWarning.ORCID_ID_MISSING,
            f'An ORCID ID is recommended."',
            node
        ))
    if not userid:
        evaluation.append((
            EvaluationWarning.USER_ID_MISSING,
            f'A User ID should be provided. An ORCID ID is recommended."',
            node
        ))
    if not email:
        evaluation.append((
            EvaluationWarning.EMAIL_MISSING,
            f'An email address should be provided."',
            node
        ))
    return evaluation


def _metadata_provider_rule(node: Node) -> list:
    return _responsible_party_rule(node)


def _other_entity_rule(node: Node) -> list:
    evaluation = []
    description = any(
        child.name == names.ENTITYDESCRIPTION and child.content
        for child in node.children
    )

    if not description:
        evaluation.append((
            EvaluationWarning.OTHER_ENTITY_DESCRIPTION_MISSING,
            f'Entity Description is highly recommended."',
            node
        ))
    return evaluation


def _personnel_rule(node: Node) -> list:
    return _responsible_party_rule(node)


def _title_rule(node: Node) -> list:
    evaluation = []
    title = node.content
    if title is not None:
        length = len(title.split(" "))
        if length < 5:
            evaluation.append((
                EvaluationWarning.TITLE_TOO_SHORT,
                f'The title "{title}" is too short. A title should have at least 5 words;'
                f' between 7 and 20 words is recommended.',
                node
            ))
    return evaluation


# ===================== End of rules section =====================


def node(node: Node):
    """
    Evaluates a given node for rule compliance.

    Args:
        node: Node instance to be evaluated

    Returns:
        None or evaluation dict
    """
    evaluation = None
    if node.name in rules:
        evaluation = rules[node.name](node)
    return evaluation


def tree(root: Node, warnings: list):
    """
    Recursively walks from the root node and evaluates
    each child node for rule compliance.

    Args:
        root: Node instance of root for evaluation
        warnings: List of warnings collected during the evaluation

    Returns:
        None
    """
    evaluation = node(root)
    if evaluation is not None:
        warnings.extend(evaluation)
    for child in root.children:
        tree(child, warnings)


# Rule function pointers
rules = {
    names.ASSOCIATEDPARTY: _associated_responsible_party_rule,
    names.CONTACT: _contact_rule,
    names.CREATOR: _creator_rule,
    names.DATASET: _dataset_rule,
    names.DATATABLE: _datatable_rule,
    names.DESCRIPTION: _description_rule,
    names.INDIVIDUALNAME: _individual_name_rule,
    names.METADATAPROVIDER: _metadata_provider_rule,
    names.OTHERENTITY: _other_entity_rule,
    names.PERSONNEL: _personnel_rule,
    names.TITLE: _title_rule,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: exceptions

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    6/5/18
"""
import daiquiri


logger = daiquiri.getLogger("exceptions: " + __name__)


class MetapypeRuleError(Exception):
    pass


class AttributeExpectedEnumError(MetapypeRuleError):
    pass


class AttributeRequiredError(MetapypeRuleError):
    pass


class AttributeUnrecognizedError(MetapypeRuleError):
    pass


class ChildNotAllowedError(MetapypeRuleError):
    pass


class ContentExpectedEmptyError(MetapypeRuleError):
    pass


class ContentExpectedEnumError(MetapypeRuleError):
    pass


class ContentExpectedFloatError(MetapypeRuleError):
    pass


class ContentExpectedNonemptyError(MetapypeRuleError):
    pass


class ContentExpectedRangeError(MetapypeRuleError):
    pass


class ContentExpectedStringError(MetapypeRuleError):
    pass


class ContentExpectedTimeFormatError(MetapypeRuleError):
    pass


class ContentExpectedDateFormatError(MetapypeRuleError):
    pass


class ContentExpectedYearFormatError(MetapypeRuleError):
    pass


class ContentExpectedUriError(MetapypeRuleError):
    pass


class MaxOccurrenceExceededError(MetapypeRuleError):
    pass


class MinOccurrenceUnmetError(MetapypeRuleError):
    pass


class StrContentUnicodeError(MetapypeRuleError):
    pass


class UnknownAttributeError(MetapypeRuleError):
    pass


class UnknownContentRuleError(MetapypeRuleError):
    pass


class UnknownNodeError(MetapypeRuleError):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: export

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    6/12/18
"""
import daiquiri

from metapype.model.node import Node
from xml.sax.saxutils import escape, unescape


logger = daiquiri.getLogger("export: " + __name__)
space = "    "


def to_xml(node: Node, level: int = 0) -> str:
    xml = ""
    closed = False
    boiler = (
        'xmlns:eml="https://eml.ecoinformatics.org/eml-2.2.0" '
        'xmlns:stmml="http://www.xml-cml.org/schema/stmml-1.2" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="https://eml.ecoinformatics.org/eml-2.2.0 '
        'https://nis.lternet.edu/schemas/EML/eml-2.2.0/xsd/eml.xsd"'
    )
    name = node.name
    attributes = ""
    for attribute in node.attributes:
        attributes += ' {0}="{1}"'.format(
            attribute, node.attributes[attribute]
        )
    if level == 0:
        indent = ""
        if name == "eml":
            name = node.name + ":" + node.name
            attributes += " " + boiler
    else:
        indent = space * level
    open_tag = "<" + name + attributes + ">"
    close_tag = "</" + name + ">"
    xml += indent + open_tag
    if node.content is not None:
        if isinstance(node.content, str):
            # if it hasn't been escaped already, escape it
            if all (x not in node.content for x in ('&amp;', '&lt;', '&gt;')):
                node.content = escape(node.content)
                # Hopefully, this is a temporary hack. Need to figure out a better way...
                # The problem is that <para> tags are treated idiosyncratically because their rules aren't fully
                #  supported. They appear within node content, unlike other tags.
                node.content = node.content.replace('&lt;para&gt;', '<para>').replace('&lt;/para&gt;', '</para>')
        xml += str(node.content) + close_tag + "\n"
        closed = True
    elif len(node.children) > 0:
        xml += "\n"
    for child in node.children:
        xml += to_xml(child, level + 1)
    if not closed:
        if len(node.children) > 0:
            xml += indent
        xml += close_tag + "\n"

    return xml


def main():
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: metapype_io

:Synopsis:
    Metapype import and export utilities

:Author:
    servilla

:Created:
    12/30/20
"""
import json
import re

import daiquiri
from lxml import etree
from xml.sax.saxutils import escape

from metapype.model.node import Node


logger = daiquiri.getLogger(__name__)


def _from_dict(node: dict, parent: Node = None) -> Node:
    """
    Build a Metapype model from a dict.

    Args:
        node: dict representation of a Metapype model
        parent: parent node of current node (root node will be None)

    Returns:
        Node: current node of Metapype model

    """
    name, body = node.popitem()
    node = Node(name, id=body[0]["id"])

    if parent is not None:
        node.parent = parent

    nsmap = body[1]["nsmap"]
    if nsmap is not None:
        for nsp in nsmap:
            node.add_namespace(nsp, nsmap[nsp])

    prefix = body[2]["prefix"]
    if prefix is not None:
        node.prefix = prefix

    attributes = body[3]["attributes"]
    if attributes is not None:
        for attribute in attributes:
            node.add_attribute(attribute, attributes[attribute])

    extras = body[4]["extras"]
    if extras is not None:
        for extra in extras:
            node.add_extras(extra, extras[extra])

    content = body[5]["content"]
    if content is not None:
        node.content = content

    tail = body[6]["tail"]
    if tail is not None:
        node.tail = tail

    children = body[7]["children"]
    for child in children:
        child_node = _from_dict(child, node)
        node.add_child(child_node)

    return node


def _format_extras(name: str, nsmap: dict) -> str:
    match = re.match(r"^\{(.*)\}(.*)$", name)
    nsname = name
    if match is not None:
        uri = match.group(1)
        target = match.group(2)
        for k, v in nsmap.items():
            if uri == v:
                nsname = f"{k}:{target}"
    return nsname


def _nsp_unique(child_nsmap: dict, parent_nsmap: dict) -> dict:
    nsmap = dict()
    for child_nsp in child_nsmap:
        if child_nsp in parent_nsmap:
            if child_nsmap[child_nsp] != parent_nsmap[child_nsp]:
                nsmap[child_nsp] = child_nsmap[child_nsp]
        else:
            nsmap[child_nsp] = child_nsmap[child_nsp]
    return nsmap


def _process_element(e, clean) -> Node:
    """
    Process an lxml etree element into a Metapype node. If the clean attribute is true, then
    remove leading and trailing whitespace from the element content.

    Args:
        e: lxml etree element
        clean: boolean to clean leading and trailing whitespace

    Returns: Node

    """
    tag = e.tag[e.tag.find("}") + 1:]  # Remove any prepended namespace

    node = Node(tag)
    node.nsmap = e.nsmap
    node.prefix = e.prefix

    if clean:
        if e.text is not None:
            node.content = None if e.text.strip() == '' else e.text.strip()
    else:
        node.content = e.text

    for name, value in e.attrib.items():
        if "{" not in name:
            node.add_attribute(name, value)
        else:
            nsname = _format_extras(name, node.nsmap)
            node.add_extras(nsname, value)

    for _ in e:
        if _.tag is not etree.Comment:
            node.add_child(_process_element(_, clean))
    for child in node.children:
        child.parent = node
        if child.nsmap == node.nsmap:
            child.nsmap = node.nsmap  # Map to single instance of nsmap
    return node


def _serialize(node: Node) -> dict:
    """
    Serializes a Metapype model instance into a Python dict

    Args:
        node: Metapype node to serialize

    Returns:
        dict: Metapype model instance dictionary

    """
    j = {node.name: []}
    j[node.name].append({"id": node.id})
    j[node.name].append({"nsmap": node.nsmap})
    j[node.name].append({"prefix": node.prefix})
    j[node.name].append({"attributes": node.attributes})
    j[node.name].append({"extras": node.extras})
    j[node.name].append({"content": node.content})
    j[node.name].append({"tail": node.tail})
    children = []
    for child in node.children:
        children.append(_serialize(child))
    j[node.name].append({"children": children})
    return j


def from_json(node: str) -> Node:
    """
    Build a Metapype model instance from JSON.

    Args:
        node: JSON Metapype model

    Returns:
        Node: current node of Metapype model

    """
    m = json.loads(node)
    return _from_dict(m)


def to_json(node: Node, indent: int = None) -> str:
    """
    Converts a Metapype model instance to JSON

    Args:
        indent:
        node: node of the model instance

    Returns:
        str: JSON of Metapype model instance

    """
    j = _serialize(node)
    return json.dumps(j, indent=indent)


def graph(node: Node, level: int = 0) -> str:
    """
    Return a graphic tree structure of the model instance

    Args:
        node: Root node of the model instance
        level: Indention level

    Returns:
        str: String representation of the model instance.
    """
    indent = "  " * level
    g = f"{node.name}[{node.id}]" if node.prefix is None else f"{node.prefix}:{node.name}[{node.id}]"
    if node.content is not None:
        g += f": {node.content}"
    if len(node.attributes) > 0:
        g += f" {str(node.attributes)}"
    if level == 0:
        g += "\n"
    else:
        g = indent + "\u2570\u2500 " + g + "\n"
    for child in node.children:
        g += graph(child, level + 1)
    return g


def from_xml(xml: str, clean: bool = True, ) -> Node:
    """
    Convert an XML model into a Metapype model. If clean is true, remove leading and trailing whitespace
    from the element content.

    Args:
        xml: XML string to be converted
        clean: boolean to clean leading and trailing whitespace

    Returns: the root Node of the Metapype model

    """
    root = _process_element(etree.fromstring(xml.encode("utf-8")), clean)
    return root


def to_xml(node: Node, parent: Node = None, level: int = 0) -> str:
    xml = ""
    indent = "  " * level

    tag = f"{node.name}" if node.prefix is None else f"{node.prefix}:{node.name}"

    attributes = ""
    if len(node.attributes) > 0:
        attributes += " ".join([f"{k}=\"{v}\"" for k, v in node.attributes.items()])

    if parent is None:
        attributes += " " + " ".join([f"xmlns:{k}=\"{v}\"" for k, v in node.nsmap.items()])
    elif node.nsmap != parent.nsmap:
        nsmap = _nsp_unique(node.nsmap, parent.nsmap)
        attributes += " " + " ".join([f"xmlns:{k}=\"{v}\"" for k, v in nsmap.items()])

    if len(node.extras) > 0:
        attributes += " " + " ".join([f"{k}=\"{v}\"" for k, v in node.extras.items()])

    if len(attributes) > 0:
        attributes = " " + attributes

    if node.content is not None:
        content = escape(node.content)
        open_tag = f"{indent}<{tag}{attributes}>{content}"
        close_tag = f"</{tag}>\n"
    else:
        open_tag = f"{indent}<{tag}{attributes}>\n"
        close_tag = f"{indent}</{tag}>\n"

    xml += open_tag
    for child in node.children:
        xml += to_xml(child, node, level + 1)
    xml += close_tag
    return xml



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: mp_io

:Synopsis:
    Utilities for reading and writing a metapype model instance

:Author:
    servilla
    costa
    ide

:Created:
    6/15/18
"""
import json

import daiquiri
import xml.etree.ElementTree as ET

from metapype.model.node import Node


logger = daiquiri.getLogger("model_io: " + __name__)

space = "    "


def from_json(json_node: dict, parent: Node = None) -> Node:
    """
    Recursively traverse Python JSON and build a metapype model
    instance.

    Args:
        json_node: JSON converted to Python structure
        parent: parent node reference to child

    Returns:
        Node: Child node of decomposed and parsed JSON

    """
    # Get first inner JSON object from dict and discard outer
    _ = json_node.popitem()
    name = _[0]
    body = _[1]
    node = Node(name, id=body[0]["id"])

    if parent is not None:
        node.parent = parent

    attributes = body[1]["attributes"]
    if attributes is not None:
        for attribute in attributes:
            node.add_attribute(attribute, attributes[attribute])

    content = body[2]["content"]
    if content is not None:
        node.content = content

    children = body[3]["children"]
    for child in children:
        child_node = from_json(child, node)
        node.add_child(child_node)

    return node


def graph(node: Node, level: int) -> str:
    """
    Return a graphic tree structure of the model instance

    Args:
        node: Root node of the model instance
        level: Indention level

    Returns:
        str: String representation of the model instance.
    """
    indent = "  " * level
    name = f"{node.name}[{node.id}]"
    if node.content is not None:
        name += ": {}".format(node.content)
    if len(node.attributes) > 0:
        name += " " + str(node.attributes)
    if level == 0:
        print(name)
    else:
        print(indent + "\u2570\u2500 " + name)
    for child in node.children:
        graph(child, level + 1)


def objectify(node: Node) -> dict:
    """
    Converts a model instance into a single Python object instance in
    preparation for JSON

    Args:
        node:

    Returns:
        dict: serialized object of the model instance

    """
    j = {node.name: []}
    j[node.name].append({"id": node.id})
    j[node.name].append({"attributes": node.attributes})
    j[node.name].append({"content": node.content})
    children = []
    for child in node.children:
        children.append(objectify(child))
    j[node.name].append({"children": children})
    return j


def to_json(node: Node):
    """
    Converts a serialized object of the model instance to a JSON compliant
    string.

    Args:
        node: Root node of the model instance

    Returns:
        str: JSON representation of the model instance

    """
    j = objectify(node)
    return json.dumps(j)


def from_xml_element(xml_elem, metapype_node, metapype_parent):
    """
    Creates a metapype node corresponding to an xml element.

    Args:
        xml_elem:  the xml element.
        metapype_node:  the metapype_node corresponding to that xml element.
                        metapype_node == None, except at the root of the tree.
        metapype_parent:  the parent metapype_node for this node.
    """
    if metapype_node is None:  # Will be None except at the root
        metapype_node = Node(name=xml_elem.tag, parent=metapype_parent)
    # xml_element_lookup_by_node_id[metapype_node.id] = (metapype_node, xml_elem)
    for name, value in xml_elem.attrib.items():
        if "}" not in name:
            metapype_node.add_attribute(name, value)
    if xml_elem.text:
        metapype_node.content = xml_elem.text
    if metapype_parent is not None:
        metapype_parent.add_child(metapype_node)
    for xml_child in xml_elem:
        from_xml_element(xml_child, None, metapype_node)


def from_xml(xml_str: str) -> Node:
    """
    Given an xml document in string form, creates the corresponding metapype model.

    Args:
        xml_str:  the xml document as a string.

    Returns:
        The root metapype node of the metapype model tree.
    """
    # Create the XML tree from text
    xml_tree = ET.ElementTree(
        ET.fromstring(_clean_namespace_expansions(_clean_xml_whitespace(xml_str)))
    )
    # Get its root
    xml_root = xml_tree.getroot()
    # Create the root node for the metapype model
    metapype_root = Node(name=xml_root.tag, parent=None)
    # Recursively build the metapype model
    from_xml_element(xml_root, metapype_root, None)
    # Return the root of the metapype model
    return metapype_root


def _clean_xml_whitespace(xml_str: str) -> str:
    """
    Given an xml document in string form, remove whitespace that is non-significant.

    Parsing xml from pretty-printed text, we will get spaces and newlines
    between elements. If an xml element's text or tail is pure whitespace,
    we assume it's not meaningful. We can't just trim all leading and
    trailing whitespace, though, because whitespace in a mixed content element
    must be assumed to be there intentionally.

    Args:
        xml_str:  the xml document as a string.

    Returns:
        The xml document as a string, with non-significant whitespace removed.
    """
    tree = ET.ElementTree(ET.fromstring(xml_str))
    xml_root = tree.getroot()
    for xml_elem in ET.ElementTree(xml_root).iter():
        if xml_elem.text and xml_elem.text.isspace():
            xml_elem.text = None
        if xml_elem.tail and xml_elem.tail.isspace():
            xml_elem.tail = None
    return ET.tostring(xml_root)


def _clean_namespace_expansions(xml_str: str) -> str:
    """
    ElementTree expands namespaces, so for example
        eml:eml
    becomes
        {eml://ecoinformatics.org/eml-2.1.1}eml
    This function strips such namespace expansions from tags.

    Args:
        xml_str:  the xml document as a string.

    Returns:
        The xml document as a string, with namespace expansions removed.
    """
    tree = ET.ElementTree(ET.fromstring(xml_str))
    xml_root = tree.getroot()
    for el in ET.ElementTree(xml_root).iter():
        if "}" in el.tag:
            el.tag = el.tag.split("}", 1)[1]  # strip all namespaces
    return ET.tostring(xml_root)


############################################################
# Everything below is work in progress. Ignore for now.

# Assumes xml_element_lookup_by_node_id dict is present.
# I.e., is intended to be called from from_xml() (currently, at least).
# def check_model_completeness() -> bool:
#
#     for _, node_tuple in xml_element_lookup_by_node_id.items():
#         metapype_node, xml_node = node_tuple
#         if metapype_node.name == 'para':
#             print(f'metapype_node.content={metapype_node.content}')
#             print(ET.tostring(xml_node).decode('utf-8'))
#             print(xml_node.text)
#         if len(metapype_node.children) != len(xml_node.getchildren()):
#             print(f'Node {metapype_node.name} is incomplete')
#             return False
#     return True


# def get_xml_as_string(node_id: str) -> str:
#     # try:
#     return ET.tostring(xml_element_lookup_by_node_id[node_id]).decode('utf-8')
